#version 430 core

in vec2 f_uv;
in vec3 f_object_normal;
in vec3 f_object_to_camera_direction;
in vec3 f_object_to_light_direction[2];
in float f_attenuation[2];
in vec3 f_normal;
in vec3 f_position;
in vec3 f_tangent_position;
in vec3 f_tangent_object_to_camera_direction;
in vec3 f_tangent_object_to_light_direction[2];

layout(location = 0) out vec4 o_color;
layout(location = 1) out vec4 o_glow_texture;

struct Light {
	vec3 world_light_position;
	vec3 light_color_ambient;
	vec3 light_color_diffuse;
	vec3 light_color_specular;
	float attenuation_constant;
	float attenuation_linear;
	float attenuation_quadratic;
	float shininess;
};

layout(location = 4) uniform sampler2D u_texture_diffuse;
layout(location = 5) uniform samplerCube u_texture_cube;
layout(location = 6) uniform vec3 u_camera_position;
layout(location = 7) uniform int instructions;
layout(location = 8) uniform Light u_lights[2]; // TODO: are 4 light sources enough?!
layout(location = 40) uniform sampler2D u_texture_normals;
layout(location = 41) uniform sampler2D u_texture_height;

// TODO: Set/get constants somewhere else
float ka = 0.2f;
float kd = 0.7f;
float ks = 0.6f;

int RENDER_TEXTURE_ONLY = 0x0001;
int GLOWS = 0x0002;
int ENVIRONMENT_MAPPING = 0x0004;
int PARALLAX_OCCLUSION_MAPPING = 0x0008;
int NO_AMBIENT = 0x0010;

// Parallax Occlusion Mapping
vec2 parallax_occlusion_mapping(vec3 view_direction, vec2 uv)
{
	const float height_scale = 0.1f;
	
	// Calculate the number of layers depending on the view direction
	const float min_layers = 8.0f;
	const float max_layers = 32.0f;
	float num_layers = mix(max_layers, min_layers, abs(dot(vec3(0.0, 0.0, 1.0), view_direction)));

	float depth_per_layer = 1.0f / num_layers;
	float current_layer_depth = 0.0f;
	vec2 current_uv = uv;
	float current_actual_depth = texture(u_texture_height, uv).r;


	// Amount to shift uv coordinates per layer along the direction of p
	vec2 p = view_direction.xy * height_scale;
	vec2 delta_uv = p / num_layers;

	while(current_layer_depth < current_actual_depth)
	{
		current_uv -= delta_uv;
		current_actual_depth = texture(u_texture_height, current_uv).r;
		current_layer_depth += depth_per_layer;
	}

	// Get depth before and after collision
	vec2 previous_uv = current_uv + delta_uv;
	float depth_before = texture(u_texture_height, previous_uv).r - current_layer_depth + depth_per_layer;
	float depth_after = current_actual_depth - current_layer_depth;

	// Interpolate uv coordinates
	float weight = depth_after / (depth_after - depth_before);
	return previous_uv * weight + current_uv * (1.0 - weight);
}

void main()
{
	// Get material colors
	vec3 material_diffuse_color = texture(u_texture_diffuse, f_uv).rgb;
	vec3 material_specular_color = clamp(2.0 * texture(u_texture_diffuse, f_uv).rgb, vec3(0.0, 0.0, 0.0), vec3(1.0, 1.0, 1.0));// TODO: texture(texture_specular, f_uv).rgb;
	vec3 material_ambient_color = material_diffuse_color * vec3(0.2, 0.3, 0.85);

	if(uint(instructions & RENDER_TEXTURE_ONLY) == RENDER_TEXTURE_ONLY)
	{
		o_color = vec4(material_diffuse_color, 1.0);
	}
	else if(uint(instructions & ENVIRONMENT_MAPPING) == ENVIRONMENT_MAPPING)
	{
		// Cube Mapping
		vec3 I = normalize(f_position - u_camera_position);
		vec3 R = reflect(I, normalize(f_normal));
		o_color = vec4(texture(u_texture_cube, R).rgb * vec3(0.9, 0.9, 0.9) * material_ambient_color * 3.0, 1.0); // make it darker!
		//o_color = vec4(texture(u_texture_cube, R).rgb * vec3(0.5, 0.5, 0.5) + ka * material_ambient_color * 2.0, 1.0); // make it darker!
		//o_color = vec4(texture(u_texture_cube, R).rgb, 1.0); // make it darker!
	}
	else
	{
		vec3 out_color_rgb = vec3(0.0, 0.0, 0.0);
		for(int i = 0; i < 2; i++)  // TODO: hardcode #lights?!
		{
			// Normalize the vectors or strange things will happen
			vec3 n = normalize(f_object_normal);
			vec3 l = normalize(f_object_to_light_direction[i]);
			vec3 v = normalize(f_object_to_camera_direction);

			// Parallax Occlusion Mapping -> change normal vector
			if(uint(instructions & PARALLAX_OCCLUSION_MAPPING) == PARALLAX_OCCLUSION_MAPPING)
			{
				v = normalize(f_tangent_object_to_camera_direction);
				vec2 new_uv = parallax_occlusion_mapping(v, f_uv);

				// TODO: maybe use this if.. depending on geometry the effect is applied to
				// Remove weird border artifacts at the edge of the parallax mapped plane
				//if(new_uv.x > 1.0 || new_uv.y > 1.0 || new_uv.x < 0.0 || new_uv.y < 0.0)
					//discard;

				// Get diffuse color and normal from new uv coordinates
				material_diffuse_color = texture(u_texture_diffuse, new_uv).rgb;
				material_specular_color = clamp(2.0 * texture(u_texture_diffuse, new_uv).rgb, vec3(0.0, 0.0, 0.0), vec3(1.0, 1.0, 1.0));
				material_ambient_color = material_diffuse_color * vec3(0.2, 0.3, 0.85);
				// Transform and normalize RGB to XYZ position
				n = normalize((2.0 * texture(u_texture_normals, new_uv) - 1.0).rgb);
				
				l = normalize(f_tangent_object_to_light_direction[i]);
			}

			// Calculating the reflection vector
			vec3 r = reflect(-l, n);

			// Calculation cos between normal and light | eye and reflection
			float cos_theta = clamp(dot(n, l), 0.0, 1.0);
			float cos_alpha = clamp(dot(v, r), 0.0, 1.0);

			// Do the shading
			vec3 color = kd * material_diffuse_color * u_lights[i].light_color_diffuse * cos_theta +
						 ks * material_specular_color * u_lights[i].light_color_specular * pow(cos_alpha, u_lights[i].shininess);

			// Apply light attenuation
			color *= f_attenuation[i];

			if(uint(instructions & NO_AMBIENT) != NO_AMBIENT){
				// Apply ambient light here because it should not be affected by light attenuation
				color += ka * material_ambient_color * u_lights[i].light_color_ambient;
			}

			// Accumulate the color
			out_color_rgb += color;
		}

		// Handle spillover
		vec3 spillover = max(vec3(0.0, 0.0, 0.0), out_color_rgb - vec3(1.0, 1.0, 1.0));
		out_color_rgb.r += spillover.g / 2.0 + spillover.b / 2.0;
		out_color_rgb.g += spillover.r / 2.0 + spillover.b / 2.0;
		out_color_rgb.b += spillover.r / 2.0 + spillover.g / 2.0;
	
		o_color = vec4(out_color_rgb, 1.0);
	}
	
	if(uint(instructions & GLOWS) == GLOWS)
	{
		o_glow_texture = texture(u_texture_diffuse, f_uv);
	}
	else
	{
		o_glow_texture = vec4(0.0, 0.0, 0.0, 1.0);
	}
}